import os

import streamlit as st


class AppState:
    """Utility class for managing state in Streamlit applications."""

    _hashmap_state_holder = {}

    @classmethod
    def _use_inmem(cls):
        return os.environ.get("STREAMLANA_USE_APP_INMEM_STATE", "").lower() in (
            "1",
            "true",
            "yes",
        )

    @classmethod
    def put(cls, key, value):
        if cls._use_inmem() or st is None:
            cls._hashmap_state_holder[key] = value
        else:
            st.session_state[key] = value

    @classmethod
    def get(cls, key, default_value=None):
        if cls._use_inmem() or st is None:
            return cls._hashmap_state_holder.get(key)
        return st.session_state.get(key, default_value)

    @classmethod
    def clear(cls, keys=None):
        if keys is None:
            keys = []
        if cls._use_inmem() or st is None:
            if keys is None:
                cls._hashmap_state_holder.clear()
            else:
                for key in keys:
                    cls._hashmap_state_holder.pop(key, None)
        else:
            if keys is None:
                for key in list(st.session_state.keys()):
                    del st.session_state[key]
            else:
                for key in keys:
                    if key in st.session_state:
                        del st.session_state[key]



STREAMLANA_ANONYMOUS_PG_NAMES = "STREAMLANA_ANONYMOUS_PG_NAMES"
STREAMLANA_URL_PG_NAME_PREFIX = "STREAMLANA_URL_PG_NAME_PREFIX"
# used for debugging
STREAMLANA_DEBUG_STATE = "STREAMLANA_DEBUG_STATE"
STREAMLANA_DEBUG_DF = "STREAMLANA_DEBUG_DF"
STREAMLANA_DEBUG_BORDER_ALL_WIDGETS = "STREAMLANA_BORDER_ALL_WIDGETS"

# used for tests
STREAMLANA_TEST_USE_APP_INMEM_STATE = "STREAMLANA_USE_APP_INMEM_STATE"



import os
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

from streamlana.app_state import AppState
from streamlana.envs import STREAMLANA_TEST_USE_APP_INMEM_STATE
from streamlana.util import get_date, substitute_placeholders

os.environ[STREAMLANA_TEST_USE_APP_INMEM_STATE] = "true"


def test_substitute_placeholders_array():
    query = "select * from table where id in __ids__"
    AppState.put("ids", [1, 2, 3, 4, 5])
    result = substitute_placeholders(query)
    assert result.strip() == "select * from table where id in (1, 2, 3, 4, 5)"


def test_substitute_placeholders():
    query = "SELECT * FROM table"
    result = substitute_placeholders(query)
    assert result == query

    query = """
    SELECT *
    FROM table
    WHERE crash_date >= '__start_date__'
      AND crash_date <= '__end_date__'
      AND status = '__status__'
    """

    AppState.put("start_date", "2024-01-01")
    AppState.put("end_date", "2024-01-31")
    AppState.put("status", "active")
    result = substitute_placeholders(query)
    expected = """
    SELECT *
    FROM table
    WHERE crash_date >= '2024-01-01'
      AND crash_date <= '2024-01-31'
      AND status = 'active'
    """
    assert result.strip() == expected.strip()


def test_get_date():
    today_dt = datetime.now(ZoneInfo("UTC")).date()
    assert get_date(None) is None
    assert get_date("today") == today_dt

    now = datetime.now(ZoneInfo("UTC"))
    expected_minus_1 = (now - timedelta(days=1)).date()
    assert get_date("today-1") == expected_minus_1



import logging
from datetime import datetime
from typing import Any, Callable, List, Optional, Sequence, Union

import pandas as pd
import streamlit as st

from streamlana.app_state import AppState
from streamlana.util import get_date


def render_rendering_failure(data: str, config_dict=None):
    """
    Render a failure message when a chart fails to render.
    nice name is it not... :)
    :param data: error message
    :param config_dict:
    :return:
    """
    if isinstance(data, str):
        msg = data
    elif isinstance(data, pd.DataFrame):
        if "message" in data.columns:
            msg = data["message"].iloc[0]
        else:
            msg = 'query does not return df with "message" column.'
    st.markdown(
        f"""
        <div style="
            display: flex;
            justify-content: center;
            align-items: center;
            border: 2px dashed #FFA726;
            border-radius: 10px;
            background-color: #fff8e1;
            text-align: center;
        ">
            <div>
                {msg}
            </div>
        </div>
        """,
        unsafe_allow_html=True,
    )


def __render_date_input(
    label: str,
    value: Union[datetime.date, str] = "Date Picker",
    min_value: Optional[datetime.date] = None,
    max_value: Optional[datetime.date] = None,
    help: Optional[str] = None,
    format: str = "YYYY/MM/DD",
    disabled: bool = False,
    label_visibility: str = "visible",
    on_change=None,
    key=None,
):
    """
    Wrapper for st.date_input with all parameters.
    """
    return st.date_input(
        label=label,
        value=value,
        min_value=min_value,
        max_value=max_value,
        help=help,
        format=format,
        disabled=disabled,
        label_visibility=label_visibility,
        on_change=on_change,
        key=key,
    )


def render_date_input(data, config_dict):
    """
    Render a date input using Streamlit's st.date_input with configuration parameters.

    Parameters:
    - data (Any): df consisting of 2 columns containing start_date and end_date respectively.
    The name of the columns should be specified in config_dict with keys 'start_date_column' and 'end_date_column'.
    If data is None or it does not contain the specified columns, the start/end dates displayed will be based on the 'value' key in config_dict.
    - config_dict (dict): Configuration dictionary containing:
        - start_date_column (str): Column name for start date in data (default is 'start_date').This can be used as placeholder in sql query as '__start_date__'  # noqa: E501
        - end_date_column (str): Column name for end date in data (default is 'end_date'). This can be used as placeholder in sql query as '__end_date__'  # noqa: E501
        - timezone (str): Timezone to use for date calculations (default is 'UTC').
        - value (str): array of 2 strings, the first is the label for start date and the second is the label for end date. ex: [today-7, today+1] or [today-7, today] or [today]  # noqa: E501
        - label (str): Label for the date input widget. default: ""
        - min_value (str): Minimum selectable date. accepted formats are 'today', 'today-N' or 'today+N' where N is an integer.  # noqa: E501
        - max_value (str): Maximum selectable date. accepted formats are 'today', 'today-N' or 'today+N' where N is an integer.  # noqa: E501
        - help (str): Optional help text.
        - format (str): Date format string. default "YYYY/MM/DD"
        - label_visibility (str): Visibility of the label ('visible', 'hidden', 'collapsed').  # noqa: E501


    """
    logging.info("Rendering Date input with configuration: %s, %s", None, data)
    start_date_col_name = config_dict.get("start_date_column", "start_date")
    end_date_col_name = config_dict.get("end_date_column", "end_date")
    timezone = config_dict.get("timezone", "UTC")

    if (
        data is not None
        and start_date_col_name in data.columns
        and start_date_col_name in data.columns
    ):
        start_date = data.loc[0, start_date_col_name].date()
        end_date = data.loc[0, end_date_col_name].date()
        date_array = [start_date, end_date]
    else:
        value = config_dict.get("value", None)
        if value is None or len(value) == 0:
            value = "[today-7, today]"
        if len(value) == 1:
            value = [value[0], value[0]]

        date_array = [
            get_date(value[0], zone=timezone),
            get_date(value[1], zone=timezone),
        ]

    configured_min_value = config_dict.get("min_value", "today-365")
    configured_max_value = config_dict.get("max_value", "today+1")
    min_date = get_date(configured_min_value, zone=timezone)
    max_date = get_date(configured_max_value, zone=timezone)
    logging.info(
        "for date_input using min_date: %s, max_date: %s, date_array: %s",
        min_date,
        max_date,
        date_array,
    )

    # AppState.put(start_date_col_name, date_array[0])
    # AppState.put(end_date_col_name, date_array[1])

    displayed_date = __render_date_input(
        label=config_dict.get("label", "Select Date Range"),
        value=date_array,
        min_value=min_date,
        max_value=max_date,
        help=config_dict.get("help", f"Timezone: {timezone}"),
        format=config_dict.get("format", "YYYY/MM/DD"),
        disabled=config_dict.get("disabled", False),
        label_visibility=config_dict.get("label_visibility", "visible"),
        key=config_dict["widget_uniq_key"],
    )
    if len(displayed_date) == 2:
        AppState.put(start_date_col_name, displayed_date[0])
        AppState.put(end_date_col_name, displayed_date[1])

    return displayed_date


def __render_json(
    body: Union[dict, list, str, Any],
    *,
    expanded: bool = True,
    width: Union[str, int] = "stretch",
):
    """
    Wrapper for st.json with all parameters.
    """
    st.json(body=body, expanded=expanded)


def render_json(df, config_dict):
    """
    Render a JSON object using Streamlit's st.json with configuration parameters.

    Parameters:
    - data (Any): The JSON data to render. expect a df with 1 column having the json.
    The column name should be specified in config_dict with key 'json_column_name'.
    - config_dict (dict): Configuration dictionary containing:
        - json_column_name: Name of the column in the df that contains the JSON object to render.
        - expanded (bool): Whether the JSON should be expanded by default.
        - width (str or int): Width of the JSON display, can be 'stretch' or an integer value.  # noqa: E501
        - body: If json_column_name is not provided, this will be used as the JSON body to render.
    """
    logging.info("Rendering JSON with configuration: %s", config_dict)
    widget_uniq_key = config_dict.get("widget_uniq_key")
    json_column_name = config_dict.get("json_column_name", None)
    if (
        json_column_name is not None
        and df is not None
        and json_column_name in df.columns
    ):
        body = df[json_column_name].iloc[0]
    else:
        body = config_dict.get("body", None)
    if body is None:
        logging.error(
            f"widget key: {widget_uniq_key}, No JSON data provided. Please provide a valid query which returns df and specify json_column_name in config_dict, or the 'body' key in config_dict."  # noqa: E501
        )
        raise ValueError(
            "No JSON data provided. Please provide a valid query which returns df with the specified json_column_name, or the 'body' key in config_dict."  # noqa: E501
        )
    logging.info("Rendering JSON with body: %s", body)
    __render_json(
        body=body,
        expanded=config_dict.get("expanded", True),
        width=config_dict.get("width", "stretch"),
    )


def render_title(df, config_dict: dict):
    """
    Render a title using Streamlit's st.title with configuration parameters.

    Parameters:
    - config_dict (dict): Configuration dictionary containing:
        - body (str): The title text.
        - anchor (str): Optional anchor for the title.
        - help (str): Optional help text.
    - df (Any): this will be ignored, but it is required to match the signature of other render functions.  # noqa: E501
    """
    logging.info("Rendering title chart with configuration: %s", config_dict)
    st.title(
        body=config_dict.get("body", "Plz set title"),
        anchor=config_dict.get("anchor", False),
        help=config_dict.get("help", "help text for title"),
        width=config_dict.get("width", "content"),
    )


def render_text(df, config_dict: dict):
    """
    Render a title using Streamlit's st.text with configuration parameters.

    Parameters:
    - config_dict (dict): Configuration dictionary containing:
        - body (str): The text.
        - help (str):
    - df (Any): this will be ignored, but it is required to match the signature of other render functions.  # noqa: E501
    """
    logging.info("Rendering text chart with configuration: %s", config_dict)
    st.text(
        body=config_dict.get("body", "Plz set text"), help=config_dict.get("help", None)
    )


def render_markdown(df, config_dict):
    """
    Render markdown using Streamlit's st.markdown with configuration parameters.

    Parameters:
    - df (Any): this will be ignored, but it is required to match the signature of other render functions.  # noqa: E501
    - config_dict (dict): Configuration dictionary containing:
        - body (str): The markdown text to render.
        - unsafe_allow_html (bool): Whether to allow HTML in the markdown.
        - help (str): Optional help text.
    """
    logging.info("Rendering markdown with configuration: %s", config_dict)
    st.markdown(
        body=config_dict.get("body", None),
        unsafe_allow_html=config_dict.get("unsafe_allow_html", False),
        help=config_dict.get("help"),
    )


def __render_dataframe(
    data: Union[pd.DataFrame, Any],
    width: Optional[int] = None,
    height: Optional[int] = None,
    *,
    use_container_width: Optional[bool] = None,
    hide_index: Optional[bool] = None,
    column_order: Optional[List[str]] = None,
    key: Optional[str] = None,
    on_select: str = "ignore",  # "ignore" | "rerun"
    selection_mode: str = "multi-row",  # "multi-row" | "single-row"
):
    """
    Wrapper for st.dataframe with all configuration options.
    todo: handle column config
    """
    st.dataframe(
        data=data,
        width=width,
        height=height,
        use_container_width=use_container_width,
        hide_index=hide_index,
        column_order=column_order,
        key=key,
        on_select=on_select,
        selection_mode=selection_mode,
    )


def render_dataframe(data, config_dict):
    """
    Render a DataFrame using Streamlit's st.dataframe with configuration parameters.

    Parameters:
    - data (pd.DataFrame): The DataFrame to render.
    - config_dict (dict): Configuration dictionary containing:
        - width (int): Width of the DataFrame. 0 means default.
        - height (int): Height of the DataFrame. 0 means default.
        - use_container_width (bool): Whether to use the full container width.
        - hide_index (bool): Whether to hide the index column.
        - column_order (list of str): Order of columns to display.
        - on_select (str): Selection behavior ('ignore' or 'rerun').
        - selection_mode (str): Selection mode ('multi-row' or 'single-row').
    """
    logging.info("Rendering dataframe chart with configuration: %s", config_dict)
    # todo: "implement column_config for dataframe."
    __render_dataframe(
        data=data,
        width=config_dict.get("width", 0),
        height=config_dict.get("height", 0),
        use_container_width=config_dict.get("use_container_width", True),
        hide_index=config_dict.get("hide_index", False),
        column_order=config_dict.get("column_order"),
        key=config_dict.get("widget_uniq_key"),
        on_select=config_dict.get("on_select", "ignore"),
        selection_mode=config_dict.get("selection_mode", "multi-row"),
    )


def render_void(void=None, config_dict: dict = None):
    """
    Render an void container.
    DEPRECATED: use render_empty()
    Parameters:
    - data (Any): will be ignored. keeping it,so its in line with other render functions
    - config_dict (dict): Configuration dictionary (optional).
    """
    st.write(" ")  # Empty string to render a void container


def render_empty(void=None, config_dict: dict = None):
    st.empty()


def render_image(df, config_dict: dict):
    st.image(
        image=config_dict.get("image"),
        caption=config_dict.get("caption"),
        width=config_dict.get("width"),
        use_column_width=config_dict.get("use_column_width", True),
        clamp=config_dict.get("clamp", False),
        channels=config_dict.get("channels", "RGB"),
        output_format=config_dict.get("output_format", "auto"),
        use_container_width=config_dict.get("use_container_width", False),
    )


def render_line_chart(data, config_dict):
    """
    Render a line chart using Streamlit's st.line_chart with configuration parameters.

    Parameters:
    - data (pd.DataFrame): The data to plot.
    - config_dict (dict): Configuration dictionary containing:
        - x (str or None): Column name for x-axis (must be in data). If None, index is used.
        - y (str, list of str, or None): Column(s) for y-axis. If None, all columns except x are used.  # noqa: E501
        - title (str): Optional title to display above the chart.
        - width (int): Width of the chart. 0 means default.
        - height (int): Height of the chart. 0 means default.
        - use_container_width (bool): Whether to use the full container width.
    """
    logging.info("Rendering line chart with configuration: %s", config_dict)
    x = config_dict.get("x")
    y = config_dict.get("y")
    title = config_dict.get("title")
    width = config_dict.get("width", 0)
    height = config_dict.get("height", 0)
    use_container_width = config_dict.get("use_container_width", True)

    __render_line_chart(
        data,
        x=x,
        y=y,
        title=title,
        width=width,
        height=height,
        use_container_width=use_container_width,
        color=config_dict.get("color", None),
    )


def __render_line_chart(
    data,
    x=None,
    y=None,
    title=None,
    width=0,
    height=0,
    use_container_width=True,
    color=None,
):
    """
    Render a line chart using Streamlit's st.line_chart.

    Parameters:
    - data (pd.DataFrame): The data to plot.
    - x (str or None): Column name to use as x-axis (must be in data). If None, index is used.  # noqa: E501
    - y (str, list of str, or None): Column(s) to plot on y-axis. If None, all columns except x are used.  # noqa: E501
    - title (str): Optional title to display above the chart.
    - width (int): Width of the chart. 0 means default.
    - height (int): Height of the chart. 0 means default.
    - use_container_width (bool): Whether to use the full container width.
    """
    if color is None:
        color = ["#fd0"]
    if title:
        st.subheader(title)

    if x:
        if y:
            plot_data = data[[x] + ([y] if isinstance(y, str) else y)].set_index(x)
        else:
            plot_data = data.set_index(x)
    else:
        plot_data = data if y is None else data[y]

    st.line_chart(
        plot_data,
        width=width,
        height=height,
        use_container_width=use_container_width,
        color=color,
    )


def __render_selectbox(
    label: str,
    options: Sequence[Any],
    index: int = 0,
    format_func: Callable[[Any], str] = str,
    key: Optional[str] = None,
    help: Optional[str] = None,
    on_change: Optional[Callable] = None,
    args: Optional[tuple] = None,
    kwargs: Optional[dict] = None,
    *,
    placeholder: Optional[str] = None,
    disabled: bool = False,
    label_visibility: str = "visible",  # "visible", "hidden", "collapsed"
) -> Any:
    """
    Full-featured wrapper for st.selectbox.
    """
    return st.selectbox(
        label=label,
        options=options,
        index=index,
        format_func=format_func,
        key=key,
        help=help,
        on_change=on_change,
        args=args,
        kwargs=kwargs,
        placeholder=placeholder,
        disabled=disabled,
        label_visibility=label_visibility,
    )


def render_selectbox(data, config_dict: dict):
    """
    Render a select box using Streamlit's st.selectbox with configuration parameters.

    Parameters:
    - data (DataFrame): Data to be used as options in the select box.
    the options will be taken from the specified column in config_dict using the column name specified by key 'options_column_name' in config_dict (defaults to 'options' if not specified). The value for 'options_column_name' be used as placeholder in sql query as '__<options_column_name>__'  # noqa: E501
    If data is None, options will be taken from config_dict['options']. Optionally, data can also contain an alias column specified by 'alias_column_name' in config_dict (defaults to 'options_column_name' if not specified).  # noqa: E501
    The alias column is used for display in the UI.
    if u have multiple select box, suggest to have different 'options_column_name' specified for each select box in the config_dict.
    - config_dict (dict): Configuration dictionary containing:
        - label (str): Label for the select box.
        - options (list): List of options to display fallback in case data is None. To pick from data, the config_dict should have the key with name of df column specified as 'options_column_name'.  # noqa: E501
        - index (int): Default selected index.
        - format_func (callable): Function to format the displayed options.
        - key (str): Unique key for the select box widget.
        - help (str): Optional help text.
        - on_change (callable): Function to call when the selection changes.
        - args (tuple): Arguments to pass to the on_change function.
        - kwargs (dict): Keyword arguments to pass to the on_change function.
        - placeholder (str): Placeholder text when no option is selected.
        - disabled (bool): Whether the select box is disabled.
        - label_visibility (str): Visibility of the label ('visible', 'hidden', 'collapsed').
        - select_box_filter_column (str): must be unique in the app. defaults to <sidebar_heading_name>_<row_idx>_<widget_idx> .This will be used as placeholder in sql query as '__<sidebar_heading_name>_<row_idx>_<widget_idx>__'  # noqa: E501
    """
    widget_uniq_key = config_dict["widget_uniq_key"]
    logging.info(
        "Rendering selectbox with configuration: %s, widget_uniq_key: %s",
        config_dict,
        widget_uniq_key,
    )

    options_column_name = config_dict.get("options_column_name", "options")
    alias_column_name = config_dict.get("alias_column_name", "options")
    if data is not None and options_column_name in data.columns:
        options = data[options_column_name].tolist()
        aliases = options
        if alias_column_name != options_column_name:
            aliases = data[alias_column_name].tolist()
    else:
        options = config_dict.get("options", [])
        aliases = options

    alias_to_option_map = dict(zip(aliases, options))

    filter_state_key = config_dict.get("options_column_name")
    filter_state_key_alias = filter_state_key + "__alias"
    filter_state_key_idx = filter_state_key + "__idx"

    if AppState.get(filter_state_key_alias) is not None:
        alias_chosen = AppState.get(filter_state_key_alias)
        if alias_chosen is not None:
            default_index = aliases.index(alias_chosen)  # noqa: F841
        else:
            default_index = config_dict.get("index", 0)  # noqa: F841
    else:
        default_index = config_dict.get("index", 0)  # noqa: F841
    # TODO : FIX THIS # noqa: F841

    alias_chosen = __render_selectbox(
        label=config_dict.get("label", "Select an option"),
        options=aliases,
        index=config_dict.get("index", 0),
        format_func=config_dict.get("format_func", str),
        key=config_dict.get("key", widget_uniq_key),
        help=config_dict.get("help", None),
        on_change=config_dict.get("on_change", None),
        args=config_dict.get("args", None),
        kwargs=config_dict.get("kwargs", None),
        placeholder=config_dict.get("placeholder", ""),
        disabled=config_dict.get("disabled", False),
        label_visibility=config_dict.get("label_visibility", "visible"),
    )
    # save the value selected
    AppState.put(filter_state_key, alias_to_option_map[alias_chosen])
    AppState.put(filter_state_key_alias, alias_chosen)
    AppState.put(filter_state_key_idx, aliases.index(alias_chosen))

    return alias_chosen


def render_metric(df, config_dict):
    """
    Render a metric using Streamlit's st.metric with configuration parameters.
    """
    logging.info("Rendering metric with configuration: %s", config_dict)
    value_column_name = config_dict.get("value_column", "value")
    delta_value_column_name = config_dict.get("delta_value_column", "delta")
    if df is not None and value_column_name in df.columns:
        value = df[value_column_name].iloc[0]
        if pd.isna(value):
            value = None
        value if isinstance(value, (str, int, float)) else str(value)
        delta = None
        if delta_value_column_name in df.columns:
            delta = df[delta_value_column_name].iloc[0]
            if pd.isna(delta):
                delta = None
    else:
        raise ValueError(
            f"Column 'Could not get metric from Metric result DataFrame. Check query for widget: {config_dict.get('widget_uniq_key')}."  # noqa: E501
        )
    st.metric(
        label=config_dict.get("label", ""),
        value=value,
        delta=delta,
        help=config_dict.get("help"),
        label_visibility=config_dict.get("label_visibility", "visible"),
        border=config_dict.get("border", False),
        width=config_dict.get("width", "content"),
    )


def render_bar_chart(data: pd.DataFrame, config_dict: dict):
    """
    Render a bar chart using Streamlit's st.bar_chart with configuration parameters.

    Parameters:
    - data (pd.DataFrame): The data to plot.
    - config_dict (dict): Configuration dictionary containing:
        - x (str or None): Column name for x-axis (must be in data). If None, index is used.  # noqa: E501
        - y (str, list of str, or None): Column(s) for y-axis. If None, all columns except x are used.
        - x_label (str): Label for the x-axis.
        - y_label (str): Label for the y-axis.
        - color (str, list of str, or None): Color(s) for the bars.
        - horizontal (bool): Whether to render the bars horizontally.
        - stack (bool or str): Stacking configuration ('normalize', 'center', 'layered', None).  # noqa: E501
        - width (int): Width of the chart. 0 means default.
        - height (int): Height of the chart. 0 means default.
        - use_container_width (bool): Whether to use the full container width.
        - title (str): Optional title to display above the chart.
    """
    logging.info("Rendering bar chart with configuration: %s", config_dict)
    __render_bar_chart(
        data=data,
        x=config_dict.get("x"),
        y=config_dict.get("y"),
        x_label=config_dict.get("x_label"),
        y_label=config_dict.get("y_label"),
        color=config_dict.get("color"),
        horizontal=config_dict.get("horizontal", False),
        stack=config_dict.get("stack", None),
        width=config_dict.get("width", 0),
        height=config_dict.get("height", 0),
        use_container_width=config_dict.get("use_container_width", True),
        title=config_dict.get("title"),
    )


def __render_bar_chart(
    data: pd.DataFrame,
    x: Optional[str] = None,
    y: Optional[Union[str, List[str]]] = None,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    color: Optional[Union[str, List[str]]] = None,
    horizontal: bool = False,
    stack: Optional[
        Union[bool, str]
    ] = None,  # Valid: True, False, "normalize", "center", "layered", None
    width: Optional[int] = None,
    height: Optional[int] = None,
    use_container_width: bool = True,
    title: Optional[str] = None,
):
    """
    Render a bar chart using Streamlit's st.bar_chart with all configuration parameters.
    """
    if title:
        st.subheader(title)

    # Don't modify the dataframe — pass x/y as-is to st.bar_chart
    st.bar_chart(
        data=data,
        x=x,
        y=y,
        x_label=x_label,
        y_label=y_label,
        color=color,
        horizontal=horizontal,
        stack=stack,
        width=width,
        height=height,
        use_container_width=use_container_width,
    )


def render_pie_chart(df, config_dict):
    """
    Render a pie chart using Plotly Express and Streamlit.
    :param df:
    :param config_dict:
        - category (str): Column name for categories (default is 'Category').
        - value (str): Column name for values (default is 'Value').
        - title (str): Title of the pie chart.
        - width (int): Width of the chart. 0 means default.
        - height (int): Height of the chart. 0 means default.
    :return:
    """
    import plotly.express as px

    fig = px.pie(
        df,
        names=config_dict.get("category", "Category"),
        values=config_dict.get("value", "Value"),
        title=config_dict.get("title", "Pie Chart"),
        color_discrete_sequence=config_dict.get("color_discrete_sequence", None),
    )
    st.plotly_chart(fig)


def render_map(data, config_dict: dict = None):
    """
    :param data:    DataFrame containing latitude and longitude columns.
    :param config_dict:
    :return:
    """
    if config_dict is None:
        config_dict = {}
    logging.info("Rendering map with configuration: %s", config_dict)
    st.map(
        data=data,
        latitude=config_dict.get("latitude"),
        longitude=config_dict.get("longitude"),
        color=config_dict.get("color"),
        size=config_dict.get("size"),
        zoom=config_dict.get("zoom", 10),
        use_container_width=config_dict.get("use_container_width", True),
        width=config_dict.get("width"),
        height=config_dict.get("height"),
    )


def render_area_chart(
    data: pd.DataFrame,
    config_dict: dict,
):
    """
    Fully parameterized wrapper for st.area_chart.
    """

    st.area_chart(
        data=data,
        x=config_dict.get("x"),
        y=config_dict.get("y"),
        x_label=config_dict.get("x_label"),
        y_label=config_dict.get("y_label"),
        color=config_dict.get("color"),
        stack=config_dict.get("stack", None),
        width=config_dict.get("width", 0),
        height=config_dict.get("height", 0),
        use_container_width=config_dict.get("use_container_width", True),
    )


def render_header(df, config_dict: dict):
    """
    Render a header using Streamlit's st.header with configuration parameters.

    Parameters:
    - config_dict (dict): Configuration dictionary containing:
        - body (str): The header text.
        - anchor (str): Optional anchor for the header.
        - help (str): Optional help text.
        - divider (bool): Whether to show a divider below the header.
    """
    logging.info("Rendering header chart with configuration: %s", config_dict)
    st.header(
        body=config_dict.get("body", "Plz set header"),
        anchor=config_dict.get("anchor", False),
        help=config_dict.get("help"),
        divider=config_dict.get("divider", False),
    )



import base64
import hashlib
import json
import logging
import os
import traceback
from importlib.resources import files
from typing import Callable, Literal

import duckdb
import streamlit as st
from streamlit.commands.page_config import InitialSideBarState, Layout

from streamlana import chart_helpers
from streamlana.envs import (
    STREAMLANA_DEBUG_DF,
    STREAMLANA_DEBUG_STATE,
    STREAMLANA_URL_PG_NAME_PREFIX,
)
from streamlana.util import substitute_placeholders


def debug_df(df):
    # todo implement
    print(df.head(5).astype(str).to_string(index=False))


def get_page_name_for_url(full_name: str, pg_anonymous=False) -> str:
    """Generate a URL-friendly name for the page.
    If the environment variable STREAMLANA_ANONYMOUS_PG_NAMES is set to "true",
    it will return a short hash of the full name.
    Otherwise, it will replace spaces and dashes with underscores and convert to lowercase.  # noqa: E501
    :param full_name: The full name of the page.
    :param pg_anonymous: If True, use a short hash for the page name.
    :return: A URL-friendly name for the page.
    """

    if pg_anonymous:
        logging.info("Using short name for page: %s", full_name)
        return short_hash(full_name)

    return full_name.replace(" ", "_").replace("-", "_").lower()


def short_hash(s, length=8):
    return hashlib.sha256(s.encode()).hexdigest()[:length]


def set_markdown_with_icon(icon_path: str = None, markdown_style: str = ""):
    """
    Set the favicon for the Streamlit application.

    :param icon_path: Path to the icon file.
    """
    if icon_path is None:
        icon_path = files("streamlana").joinpath("streamlana.ico")
    with open(icon_path, "rb") as icon_file:
        encoded = base64.b64encode(icon_file.read()).decode()
    favicon_data_url = f"data:image/x-icon;base64,{encoded}"
    st.markdown(
        markdown_style
        + f"""
        <link rel="icon" href="{favicon_data_url}">
        <script>
          var link = document.querySelector("link[rel~='icon']");
          if (!link) {{
            link = document.createElement('link');
            link.rel = 'icon';
            document.getElementsByTagName('head')[0].appendChild(link);
          }}
          link.href = "{favicon_data_url}";
        </script>
        """,
        unsafe_allow_html=True,
    )


def auth_enabled(auth_session_state_key_name) -> bool:
    """
    Check if authentication is enabled based on session state key.

    :param auth_session_state_key_name: Key name in session state to check for authentication.  # noqa: E501
    :return: True if authentication is enabled, False otherwise.
    """
    return auth_session_state_key_name is not None


def load_side_bar_config_yaml(side_bar_config_file_path):
    """
    Load sidebar configuration from a YAML file.

    :param side_bar_config_file_path: Path to the YAML file containing sidebar configuration.  # noqa: E501
    :return: Parsed sidebar configuration as a list of dictionaries.
    """
    import yaml

    try:
        with open(side_bar_config_file_path, "r") as f:
            config = yaml.safe_load(f)
    except Exception as e:
        logging.error(
            "Failed to load sidebar configuration from file: %s",
            side_bar_config_file_path,
        )
        traceback.print_exc()
        raise e
    return config.get("side_bar", [])


def __create_dynamic_func(
    full_name: str, config_dict, duckbdb_conn, pg_anonymous=False
):
    name_in_url = get_page_name_for_url(full_name, pg_anonymous=pg_anonymous)

    prefix = os.environ.get(STREAMLANA_URL_PG_NAME_PREFIX, "pg_")
    func_name = f"{prefix}{name_in_url}"

    def _template():
        render_dashboard(full_name, config_dict, duckbdb_conn=duckbdb_conn)

    _template.__name__ = func_name  # Change the function’s __name__ attribute
    globals()[func_name] = _template  # Inject into global namespace

    return globals()[func_name]


def default_check_user_access() -> str:
    """Check if the user has access to pages.
    :param Any: Placeholder for any input
    :return: None if not authorized, Username if allowed (return "guest" if auth disabled)  # noqa: E501
    """
    # Assuming that auth is disabled.
    return "guest"


def render_side_bar_pages(
    side_bar_config: dict,
    duckbdb_conn,
    markdown_style: str = "",
    logo_path: str = None,
    logo_size: Literal["small", "medium", "large"] = "large",
    icon=None,
    check_user_access: Callable[[], str] = default_check_user_access,
):
    """Render sidebar pages based on the provided configuration.
    :param side_bar_config: Configuration for sidebar pages.
    :param duckbdb_conn: DuckDB connection object.
    :param markdown_style: Custom CSS for sidebar links.
    :param logo_path: Path to the logo image.
    :param logo_size: Size of the logo. small, medium, large.
    :param icon: Path to the icon file for the favicon.
    :param check_user_access: Function to check user access and return username.
    """
    # set_page_layout(
    #     initial_sidebar_state="expanded",
    #     layout="wide",
    #     page_icon="streamlana.ico",
    # )
    username = "guest"
    if check_user_access is not None:
        username = check_user_access()
        if username is None:
            logging.warning(
                "check_user_access returned 'None' username. User is not allowed to access pages. Returning without rendering sidebar pages."  # noqa: E501
            )
            return
    else:
        logging.warning(
            "check_user_access callback is None. will assume auth is disabled. username set to guest."  # noqa: E501
        )

    logging.info("username obtained from check_user_access callabck: %s", username)
    # Track selected file in session state
    if "selected_page" not in st.session_state:
        st.session_state.selected_page = None

    set_markdown_with_icon(icon_path=icon, markdown_style=markdown_style)

    if logo_path == "" or logo_path is None:
        logo_path = files("streamlana").joinpath("streamlana.png")

    st.logo(logo_path, size=logo_size)

    # Create a dictionary to hold pages
    pages = {}
    for section in side_bar_config:
        ordered_pages = []
        heading = section.get("heading", "")
        if heading == "":
            raise ValueError(
                "heading is required for each section in side_bar_config. "
                "Please provide a valid heading."
            )
        page_sections = section.get("pages", None)
        if pages is None:
            raise ValueError(
                f"pages are required for section: '{heading}'. "
                "Please provide a valid list of pages."
            )
        for pg_section in page_sections:
            name = pg_section.get("name", "")
            pg_enabled = pg_section.get("enabled", True)
            pg_anonymous = pg_section.get("anonymous", False)
            if not pg_enabled:
                logging.info(
                    f"Page '{name}' in section '{heading}' is disabled, skipping rendering."  # noqa: E501
                )
                continue
            if name.strip() == "":
                raise ValueError(
                    f"name is required for each page in section: '{heading}'. "
                    "Please provide a valid name."
                )

            path_to_py = pg_section.get("code_definition", None)
            if path_to_py is not None:
                url_path = get_page_name_for_url(name, pg_anonymous=pg_anonymous)
                ordered_pages.append(st.Page(path_to_py, title=name, url_path=url_path))
            else:
                config_file_path = pg_section.get("config_file_path")
                if config_file_path is None:
                    raise ValueError(
                        f"since path to custom python module not specified for page, page config file path is required for page: '{name}'"  # noqa: E501
                    )
                try:
                    with open(config_file_path, "r") as f:
                        config_dict = json.load(f)
                        callable_render = __create_dynamic_func(
                            name, config_dict, duckbdb_conn, pg_anonymous=pg_anonymous
                        )
                        ordered_pages.append(st.Page(callable_render, title=name))
                except Exception as e:
                    logging.error("error reading the config file: %s", config_file_path)
                    traceback.print_exc()
                    st.error("⚠ Error reading config file: %s" % config_file_path)
                    raise e

        pages[heading] = ordered_pages
    pg = st.navigation(pages, position="sidebar", expanded=True)
    pg.run()
    if os.environ.get(STREAMLANA_DEBUG_STATE, "false").lower() == "true":
        st.sidebar.subheader("Session State Debug")
        st.sidebar.json(st.session_state.to_dict())


def set_page_layout(
    page_title="StreamLana",
    page_icon=None,
    layout: Layout = "wide",
    initial_sidebar_state: InitialSideBarState = "expanded",
):
    """
    Set the  Length of Width distribution arraypage layout for Streamlit application.
    call it only once. at the top of your application.
    """
    if page_icon is None:
        page_icon = files("streamlana").joinpath("streamlana.ico")

    if "_layout" not in st.session_state:
        st.session_state._layout = layout
    st.set_page_config(
        page_title=page_title,
        page_icon=page_icon,
        layout=st.session_state._layout,  # Optional: "centered" or "wide"
        initial_sidebar_state=initial_sidebar_state,  # Optional: "expanded" or "collapsed",  # noqa: E501
        menu_items={
            "Get help": "https://github.com/jaihind213/streamlana",
            "About": "Streamlit in Grafana style. Business Intelligence as Configuration",  # noqa: E501
        },
    )


def render_dashboard(dashboard_name: str, page_config_dict, duckbdb_conn=None):
    title = page_config_dict.get("page_title", "")
    if title is not None and title.strip() != "":
        st.title(title)

    page_rows = page_config_dict.get("page_rows", [])
    num_rows_in_page = len(page_rows)
    logging.info(
        "Rendering dashboard: %s with %d rows", dashboard_name, num_rows_in_page
    )
    for row_idx, row in enumerate(page_rows):
        widgets_width_distribution = row.get("widgets_width_spec", [])
        with_expander = row.get(
            "with_expander", {"label": f"Row {row_idx} Widgets...", "expanded": False}
        )
        widgets_gap = row.get("widgets_gap", "small")
        widgets_v_alignment = row.get("widgets_vertical_alignment", "top")
        widgets_have_border = row.get("widgets_border", False)
        widgets = row.get("widgets", [])
        if len(widgets) != len(widgets_width_distribution):
            raise ValueError(
                f"Length of 'widgets_width_spec' array: {len(widgets_width_distribution)}, does not match number of widgets: len(widgets) in row:{row_idx} of page '{dashboard_name}'."  # noqa: E501
            )
        logging.info(
            "Rendering row in dashboard: %s ,with width distribution: %s",
            dashboard_name,
            widgets_width_distribution,
        )
        if with_expander is not None:
            with st.expander(
                label=with_expander.get("label", ""),
                expanded=with_expander.get("expanded", False),
            ):
                cols = st.columns(
                    widgets_width_distribution,
                    gap=widgets_gap,
                    vertical_alignment=widgets_v_alignment,
                    border=widgets_have_border,
                )
        else:
            cols = st.columns(
                widgets_width_distribution,
                gap=widgets_gap,
                vertical_alignment=widgets_v_alignment,
                border=widgets_have_border,
            )

        logging.info(
            "Rendering dashboard: %s with %d widgets", dashboard_name, len(widgets)
        )
        for widget_idx, widget in enumerate(widgets):
            widget_enabled = widget.get("widget_enabled", True)
            if not widget_enabled:
                logging.info(
                    f"Widget at index {widget_idx} in row {row_idx} is disabled, skipping rendering."  # noqa: E501
                )
                chart_helpers.render_void()
                continue
            query = widget.get("query", "select 1 as dummy_result")
            config = widget.get("config", {})
            config["widget_uniq_key"] = f"{dashboard_name}_{row_idx}_{widget_idx}"
            widget_type = widget.get("type", None)
            if widget_type is None or "" == widget_type.strip():
                raise ValueError(
                    f"Widget type is required for widget at index {widget_idx} in row {row_idx}, dashboard: {dashboard_name}"  # noqa: E501
                )

            # Run the query to get a dataframe
            final_query = substitute_placeholders(query)
            logging.info(
                f"Executing final query: {final_query} for widget type: {widget_type} with config: {config}, page_row: {row_idx}, widget_idx: {widget_idx}, dashboard: {dashboard_name}"  # noqa: E501
            )
            # Execute the query
            if duckbdb_conn is None:
                logging.warning(
                    "DuckDB connection is not provided, creating a new connection."
                )
                duckbdb_conn = duckdb.connect()
            try:
                df = duckbdb_conn.execute(final_query).df()
                if os.environ.get(STREAMLANA_DEBUG_DF, "false") == "true":
                    debug_df(df)
            except Exception:
                logging.error(
                    f"Error executing query: {final_query} for widget type: {widget_type}, dashboard: {dashboard_name}, row: {row_idx}, widget_idx: {widget_idx}"  # noqa: E501
                )
                traceback.print_exc()
                with cols[widget_idx]:
                    chart_helpers.render_rendering_failure(
                        "⚠ Query Failed.\n\n check logs", None
                    )
                continue

            # Dispatch to the appropriate render function
            render_fn = getattr(chart_helpers, f"render_{widget_type}", None)

            if callable(render_fn):
                try:
                    with cols[widget_idx]:
                        render_fn(df, config)
                except Exception:
                    logging.error(
                        f"Error rendering widget type: {widget_type}, dashboard: {dashboard_name}, row: {row_idx}, widget_idx: {widget_idx}"  # noqa: E501
                    )
                    traceback.print_exc()
                    with cols[widget_idx]:
                        chart_helpers.render_rendering_failure(
                            "⚠ Render Error.\n\n Check Logs", None
                        )
                    continue
            else:
                with cols[widget_idx]:
                    st.error(
                        f"Unsupported widget type: {widget_type}, dashboard: {dashboard_name}, row: {row_idx}, widget_idx: {widget_idx}"  # noqa: E501
                    )



import logging
import re
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import numpy as np

from streamlana.app_state import AppState


def get_date(value, zone="UTC"):
    if value is None:
        return None
    value = value.lower().strip()

    if value == "today":
        return datetime.now(ZoneInfo(zone)).date()
    elif "today" in value and "-" in value:
        try:
            days_ago = int(value.split("-")[1])
            to_return = (datetime.now(ZoneInfo(zone)) - timedelta(days=days_ago)).date()
            return to_return
        except (IndexError, ValueError):
            raise ValueError(
                "Invalid format. Use 'today' or 'today-N/today+N' where N is an integer."  # noqa: E501
            )
    elif "today" in value and "+" in value:
        try:
            days_ahead = int(value.split("+")[1])
            to_return = (
                datetime.now(ZoneInfo(zone)) + timedelta(days=days_ahead)
            ).date()
            return to_return
        except (IndexError, ValueError):
            raise ValueError(
                "Invalid format. Use 'today' or 'today-N/today+N' where N is an integer."  # noqa: E501
            )
    else:
        raise ValueError("Unsupported value. Use 'today' or 'today-N' format.")


def substitute_placeholders(query):
    """
    Finds all placeholders in __ __ and replaces each with values from FilterState.
    :param query:
    :return: final query
    """

    def replacer(match):
        key = match.group(1)
        val = AppState.get(key)
        final_value = val
        if val is None:
            raise KeyError(
                f"Missing key in FilterState: {key}. the widget may not have been rendered yet. Please ensure that the widget is rendered before using its value in a query."  # noqa: E501
            )
        if isinstance(val, (list, tuple, np.ndarray)):
            final_value = tuple(val)
            if all(isinstance(x, bool) for x in final_value) or all(
                isinstance(x, np.bool) for x in final_value
            ):
                final_value = tuple("TRUE" if x else "FALSE" for x in final_value)
        return f"{final_value}" if isinstance(final_value, str) else str(final_value)

    final_query = re.sub(r"__([a-zA-Z0-9_]+)__", replacer, query)
    logging.info(f"Final query after substitution: {final_query}")
    return final_query
